from datetime import datetime, timedelta
from typing import Dict, List, Optional, Tuple
from laboris.ui import Table, error
from laboris.util import date_parser as date, taskQuery
from pony import orm
from halo import Halo
import click

# __BAR_CHARS__ = ["█", "▉", "▊", "▋", "▌", "▍", "▎", "▏"]
__BAR_CHARS__ = ["█"]
__BAR_WIDTH__ = 120


@click.command(name="gantt")
@click.option(
    "--hidden/--no-hidden", default=False, help="Include hidden tasks in the results"
)
@click.option(
    "-s",
    "--start",
    type=date,
    default=datetime.now() - timedelta(days=7),
    help="Set the starting date for the report",
)
@click.option(
    "-e",
    "--end",
    type=date,
    default=datetime.now() + timedelta(hours=1),
    help="Set the ending date for the report",
)
@click.option(
    "-S",
    "--day-start",
    type=int,
    default=0,
    help="Set what out of the day to start the graph at",
)
@click.option(
    "-E",
    "--day-end",
    type=int,
    default=23,
    help="Set what out of the day to end the graph at",
)
@click.argument("query", nargs=-1)
def gantt_command(
    query: Tuple[str, ...],
    start: datetime,
    end: datetime,
    day_start: int,
    day_end: int,
    hidden: bool,
):
    """Generate a graph displaying time spent activly working on every task"""

    if start >= end:
        click.echo(error(f"Start date must be before end date {start} >= {end}"))
        return 1
    elif day_start >= day_end:
        click.echo(
            error(f"Day start hour must be before the end hour {start} >= {end}")
        )
        return 1

    with orm.db_session:
        spinner = Halo(text="Constructing report")
        spinner.start()
        tasks = list(
            taskQuery(
                lambda t: (hidden or t.hidden == False)
                and (t.createdDate >= start or t.doneDate >= start)
                and (t.createdDate <= end or t.doneDate <= end),
                *query,
            )
        )

        keys: List[str] = []
        data: Dict[str, Tuple[datetime, List[Tuple[datetime, str]]]] = {}

        multi: Dict[str, bool] = {
            "year": start.year != end.year,
            "month": start.month != end.month,
        }
        fmt = "%b %d %Y" if multi["year"] else ("%b %d" if multi["month"] else "%d")

        current = start
        while current <= end:
            key = current.strftime(fmt)
            data[key] = (current.replace(hour=0, minute=0, second=0, microsecond=0), [])
            keys.append(key)
            current += timedelta(days=1)
        key = end.strftime(fmt)
        if key not in data:
            data[key] = (current.replace(hour=0, minute=0, second=0, microsecond=0), [])
            keys.append(key)

        for task in tasks:
            color = task.color.strip("#")
            for ev in task.events:
                estart = datetime.fromtimestamp(ev[0])
                if estart.strftime(fmt) in keys:
                    data[estart.strftime(fmt)][1].append((estart, color))
                eend = datetime.fromtimestamp(ev[1]) if len(ev) == 2 else datetime.now()
                if eend.strftime(fmt) in keys:
                    data[eend.strftime(fmt)][1].append((eend, color))

        width = click.get_terminal_size()[0] - max(len(x) for x in data.keys()) + 2

        x_axis = " " * int(
            (width - ((day_end - day_start + 1) * 2)) // (day_end - day_start + 1)
        )
        x_axis = x_axis.join(f"{x:<2}" for x in range(day_start, day_end + 1))

        actual_width = len(x_axis)
        conversion = (
            actual_width
            * len(__BAR_CHARS__)
            / timedelta(hours=(day_end - day_start)).total_seconds()
        )

        table = Table()

        color_stack: List[Optional[str]] = [None]
        for key in keys:
            head = data[key][0] + timedelta(hours=day_start)
            events = sorted(data[key][1], key=lambda x: x[0])

            cells: List[Optional[str]] = [None] * (actual_width * len(__BAR_CHARS__))

            iprev = 0
            for ev in events:
                istart = int((ev[0] - head).total_seconds() * conversion)
                if istart < 0:
                    istart = 0
                elif istart >= len(cells):
                    istart = len(cells) - 1
                for i in range(iprev, istart):
                    cells[i] = color_stack[-1]
                iprev = istart
                if len(color_stack) > 1 and color_stack[-1] == ev[1]:
                    color_stack.pop(-1)
                elif len(color_stack) == 1 and color_stack[0] == ev[1]:
                    color_stack.pop(-1)
                else:
                    color_stack.append(ev[1])

            day: List[Tuple[str, Optional[str], Optional[str]]] = []
            for i in range(actual_width):
                block = cells[i * len(__BAR_CHARS__) : (i + 1) * len(__BAR_CHARS__)]
                day.append(
                    (
                        __BAR_CHARS__[
                            len(__BAR_CHARS__) - sum(1 for x in block if x == block[0])
                        ],
                        block[0],
                        block[-1],
                    )
                )

            row = ""
            for ch, fg, bg in day:
                # if fg and bg:
                #     row += click.style(
                #         ch,
                #         fg=tuple(int(fg[i : i + 2], 16) for i in (0, 2, 4)),
                #         bg=tuple(int(bg[i : i + 2], 16) for i in (0, 2, 4)),
                #     )
                # elif bg:
                #     row += click.style(
                #         ch,
                #         fg="black",
                #         bg=tuple(int(bg[i : i + 2], 16) for i in (0, 2, 4)),
                #     )
                # elif fg:
                #     row += click.style(
                #         ch,
                #         fg=tuple(int(fg[i : i + 2], 16) for i in (0, 2, 4)),
                #         bg="black",
                #     )
                # else:
                #     row += click.style(" ", bg="black")
                if fg and bg:
                    row += click.style(
                        ch,
                        fg=tuple(int(fg[i : i + 2], 16) for i in (0, 2, 4)),
                        bg=tuple(int(bg[i : i + 2], 16) for i in (0, 2, 4)),
                    )
                elif bg:
                    row += click.style(
                        ch,
                        bg=tuple(int(bg[i : i + 2], 16) for i in (0, 2, 4)),
                    )
                elif fg:
                    row += click.style(
                        ch,
                        fg=tuple(int(fg[i : i + 2], 16) for i in (0, 2, 4)),
                    )
                else:
                    row += click.style(" ")

            table.append([click.style(key, bold=True), row])

        table.append(["", x_axis])

        spinner.stop()

        click.echo(table.format())



from datetime import datetime, timedelta
from laboris.ui import Table
from laboris.util import pick_one_task
from pony import orm
import humanize

import click


@click.command(name="detail")
@click.option("--csv/--no-csv", default=False, help="Format output as a CSV")
@click.argument("label", nargs=1)
def detail_command(label: str, csv: bool):
    """List details of a specific task"""
    with orm.db_session:
        db_task = pick_one_task(label)
        if db_task is None:
            return 1

        table = Table(
            header=["Key", "Value"],
            header_style={"underline": True, "bold": True},
            align=[">", "<"],
        )
        table.append(
            [
                click.style("UUID", bold=True),
                click.style(str(db_task.uuid), fg="bright_black"),
            ]
        )
        table.append(
            [click.style("State", bold=True), click.style(db_task.state, fg="cyan")]
        )
        table.append([click.style("Label", bold=True), db_task.label])
        table.append(
            [
                click.style("Priority", bold=True),
                click.style(db_task.priority, **db_task.priorityColor),
            ]
        )

        tag_table = Table(header=["UUID", "Label"], header_style={"bold": True})
        for tag in db_task.tags:
            tag_table.append(
                [
                    click.style(tag.uuid, fg="bright_black"),
                    click.style(tag.label, fg="yellow"),
                ]
            )
        table.append(
            [
                click.style("Tags", bold=True),
                tag_table.format() if len(db_task.tags) != 0 else "",
            ]
        )

        parent_table = Table(header=["State", "UUID", "Label"], header_style={"bold": True})
        for task in db_task.parents:
            parent_table.append(
                [
                    click.style(task.state, fg="cyan"),
                    click.style(task.uuid, fg="bright_black"),
                    click.style(task.label, fg="blue"),
                ]
            )
        table.append(
            [
                click.style("Parent Tasks", bold=True),
                parent_table.format() if len(db_task.parents) else "",
            ]
        )

        child_table = Table(header=["State", "UUID", "Label"], header_style={"bold": True})
        for task in db_task.children:
            child_table.append(
                [
                    click.style(task.state, fg="cyan"),
                    click.style(task.uuid, fg="bright_black"),
                    click.style(task.label, fg="blue"),
                ]
            )
        table.append(
            [
                click.style("Child Tasks", bold=True),
                child_table.format() if len(db_task.children) else "",
            ]
        )

        table.append(
            [
                click.style("Due Date", bold=True),
                click.style(
                    f"{db_task.dueDate:%c}" if db_task.dueDate else "",
                    fg="magenta",
                ),
            ]
        )

        table.append(
            [
                click.style("Goal", bold=True),
                click.style(
                    f"{humanize.naturaldelta(db_task.goalDuration)} / {humanize.naturaldelta(db_task.goalPeriod)}"
                    if db_task.goalDuration and db_task.goalPeriod
                    else "",
                    fg="magenta",
                ),
            ]
        )

        table.append(
            [
                click.style("Created Date", bold=True),
                click.style(f"{db_task.createdDate:%c}", fg="yellow"),
            ]
        )
        table.append(
            [
                click.style("Updated Date", bold=True),
                click.style(f"{db_task.updatedDate:%c}", fg="yellow"),
            ]
        )
        table.append(
            [
                click.style("Done Date", bold=True),
                click.style(
                    f"{db_task.doneDate:%c}" if db_task.doneDate else "", fg="yellow"
                ),
            ]
        )

        event_table = Table(
            header=["Start", "Stop", "Duration"],
            header_style={"bold": True},
            align=["<", "<", "<"],
        )
        for event in db_task.events:
            ev = [datetime.fromtimestamp(x) for x in event]
            if len(ev) == 2:
                event_table.append(
                    [
                        click.style(f"{ev[0]:%c}", fg="yellow"),
                        click.style(f"{ev[1]:%c}", fg="yellow"),
                        click.style(f"{ev[1] - ev[0]:}", fg="green"),
                    ]
                )
            else:
                event_table.append(
                    [
                        click.style(f"{ev[0]:%c}", fg="yellow"),
                        click.style("NOW", fg="yellow", bold=True),
                        click.style(f"{datetime.now() - ev[0]}", fg="green"),
                    ]
                )

        if len(db_task.events) != 0:
            total = timedelta()
            for event in db_task.events:
                if len(event) == 2:
                    total += timedelta(seconds=event[1] - event[0])
                else:
                    total += timedelta(seconds=datetime.now().timestamp() - event[0])
            event_table.append(["", "", click.style(f"{total}", fg="green", bold=True)])

        table.append(
            [
                click.style("Events", bold=True),
                event_table.format() if len(db_task.events) else "",
            ]
        )

        table.append(
            [
                click.style("Urgency", bold=True),
                click.style(f"{db_task.urgency:4.1f}", **db_task.urgencyColor),
            ]
        )

        click.echo(table.format(csv))



from laboris.util import pick_one_task
from pony import orm

import click

from laboris.model import Task
from laboris.ui import task_short


@click.command(name="delete")
@click.argument("label", nargs=1)
def delete_command(
    label: str,
):
    """Delete an existing task"""
    with orm.db_session:
        db_task = pick_one_task(label)
        if db_task is None:
            return 1
        elif click.confirm(f"Do you want to delete {task_short(db_task)}?", abort=True):
            rep = task_short(db_task)
            db_task.delete()
            click.echo(f"Deleted the task {rep}")



from typing import Optional
from laboris.util import date_parser as date, pick_one_task
from pony import orm
from datetime import datetime

import click

from laboris.ui import error, task_short


@click.command(name="stop", context_settings={"ignore_unknown_options": True})
@click.argument("label", nargs=1)
@click.argument("date", type=date, required=False)
def stop_command(label: str, date: Optional[datetime]):
    """Stop tracking activity on a task"""
    with orm.db_session:
        db_task = pick_one_task(label)
        if db_task is None:
            return 1
        elif len(db_task.events) == 0 or len(db_task.events[-1]) != 1:
            click.echo(
                error(f"The task {task_short(db_task)}")
                + error(" is not currently active")
            )
            return 1
        db_task.events[-1].append(
            date.timestamp() if date else datetime.now().timestamp()
        )
        db_task.updatedDate = datetime.now()
        click.echo(f"Stopped progress on {task_short(db_task)}")



from datetime import datetime, timedelta
from typing import Callable, List, Optional, Union
from pony import orm
import click
import re
import isodate
from dateutil import parser

from laboris.model import Tag, Task
from laboris.ui import error

__DURATION_PARSER__ = re.compile(
    r"((?P<year>[0-9]+)y)?((?P<month>[0-9]+)M)?((?P<week>[0-9]+)w)?((?P<day>[0-9]+)d)?((?P<hour>[0-9]+)h)?((?P<minute>[0-9]+)m)?((?P<second>[0-9]+)s)?"
)
__RELATIVE_DURATION_PARSER__ = re.compile(
    r"(?P<sign>[+-])?((?P<year>[0-9]+)y)?((?P<month>[0-9]+)M)?((?P<week>[0-9]+)w)?((?P<day>[0-9]+)d)?((?P<hour>[0-9]+)h)?((?P<minute>[0-9]+)m)?((?P<second>[0-9]+)s)?"
)
__CONDITION_REGEX__ = re.compile(
    r"(?P<lhs>[^!~=<>]+)(?P<op>[~~=<>]{1,2})(?P<rhs>[^!~=<>]+)"
)


def date_parser(string: Union[str, datetime]) -> datetime:
    if isinstance(string, datetime):
        return string
    match = re.match(__RELATIVE_DURATION_PARSER__, string)
    if match is None or not match.group(0):
        return parser.parse(string)
    else:
        now = datetime.now()
        sign = -1 if match.group("sign") == "-" else 1
        if match.group("year"):
            now += timedelta(days=sign * 365 * int(match.group("year")))
        if match.group("month"):
            now += timedelta(days=sign * 30 * int(match.group("month")))
        if match.group("week"):
            now += timedelta(days=sign * 7 * int(match.group("week")))
        if match.group("day"):
            now += timedelta(days=sign * int(match.group("day")))
        if match.group("hour"):
            now += timedelta(hours=sign * int(match.group("hour")))
        if match.group("minute"):
            now += timedelta(minutes=sign * int(match.group("minute")))
        if match.group("second"):
            now += timedelta(seconds=sign * int(match.group("second")))
        return now


def duration_parser(string: Union[str, timedelta]) -> timedelta:
    if isinstance(string, timedelta):
        return string
    match = re.match(__DURATION_PARSER__, string)
    if match is None or not match.group(0):
        return isodate.parse_duration(string)
    else:
        now = timedelta()
        if match.group("year"):
            now += timedelta(days=365 * int(match.group("year")))
        if match.group("month"):
            now += timedelta(days=30 * int(match.group("month")))
        if match.group("week"):
            now += timedelta(days=7 * int(match.group("week")))
        if match.group("day"):
            now += timedelta(days=int(match.group("day")))
        if match.group("hour"):
            now += timedelta(hours=int(match.group("hour")))
        if match.group("minute"):
            now += timedelta(minutes=int(match.group("minute")))
        if match.group("second"):
            now += timedelta(seconds=int(match.group("second")))
        return now


def pick_one_task(label: str, state: List[str] = ["OPEN"]) -> Optional[Task]:
    tasks = list(
        Task.select(lambda t: t.state in state and t.label.lower().startswith(label))
    )

    if len(tasks) == 1:
        return tasks[0]

    tasks = list(Task.select(lambda t: t.state in state))
    tasks = [x for x in tasks if str(x.uuid).startswith(label)]

    if len(tasks) == 0:
        click.echo(error(f"Faild to find a task with the label {repr(label)}"))
        return None
    elif len(tasks) != 1:
        click.echo(error(f"Found multiple tasks matching the label {repr(label)}"))
        return None
    return tasks[0]


def taskQuery(init: Callable, *args):
    query = Task.select(init)

    for arg in args:
        if type(arg) != str:
            query = query.where(arg)
        elif arg[0] == "@":
            tag = Tag.select(lambda x: x.label == arg[1:]).get()
            query = query.where(lambda t: tag in t.tags)
        elif arg[0] == "+":
            task = Task.select(lambda x: x.label == arg[1:]).get()
            query = query.where(lambda t: task in t.parents)
        elif arg[0] == "_":
            task = Task.select(lambda x: x.label == arg[1:]).get()
            query = query.where(lambda t: task in t.children)
        else:
            m = re.match(__CONDITION_REGEX__, arg)
            if m is None:
                query = query.where(lambda t: arg in t.label)
                continue

            lhs = m.group("lhs")
            op = m.group("op")
            rhs = m.group("rhs")

            if op == "~=":
                op = "in"

            lhs_key = False
            if lhs in (
                "state",
                "hidden",
                "label",
                "priority",
                "dueDate",
                "goalDuration",
                "goalPeriod",
                "createdDate",
                "updatedDate",
                "doneDate",
            ):
                lhs_key = True

            key = lhs if lhs_key else rhs
            value = rhs if lhs_key else lhs

            if key in (
                "dueDate",
                "createdDate",
                "updatedDate",
                "doneDate",
            ):
                value = date_parser(value)
            elif key in ("priority",):
                value = int(value)
            elif key in ("goalDuration", "goalPeriod"):
                value = duration_parser(value)
            elif key in ("hidden"):
                value = True if value.lower() == "true" else False

            query = query.where(
                f"{'t.' + key if lhs_key else 'x'} {op} {'x' if lhs_key else 't.' + key}",
                {"x": value},
            )

    return query



from laboris.util import pick_one_task
from pony import orm
from datetime import datetime

import click

from laboris.ui import task_short


@click.command(name="done")
@click.argument("label", nargs=1)
def done_command(label: str):
    """Mark a task as completed"""
    with orm.db_session:
        db_task = pick_one_task(label)
        if db_task is None:
            return 1

        if len(db_task.events) != 0 and len(db_task.events[-1]) == 1:
            db_task.events[-1].append(datetime.now().timestamp())
        db_task.updatedDate = datetime.now()
        db_task.doneDate = datetime.now()
        db_task.state = "DONE"
        click.echo(f"Marked the task {task_short(db_task)} as completed")



from datetime import datetime, timedelta
from typing import Dict, List, Tuple
from laboris.ui import Table, error
from laboris.util import date_parser as date, taskQuery
from pony import orm
from halo import Halo
import click

__BAR_CHAR__ = "■"
__BAR_WIDTH__ = 40


@click.command(name="graph")
@click.option(
    "--hidden/--no-hidden", default=False, help="Include hidden tasks in the results"
)
@click.option(
    "-s",
    "--start",
    type=date,
    default=datetime.now() - timedelta(days=7),
    help="Set the starting date for the report",
)
@click.option(
    "-e",
    "--end",
    type=date,
    default=datetime.now() + timedelta(hours=1),
    help="Set the ending date for the report",
)
@click.option(
    "-p",
    "--period",
    default="day",
    type=click.Choice(["year", "month", "week", "day"], case_sensitive=False),
    help="Set the bin size for the graph",
)
@click.argument("query", nargs=-1)
def graph_command(
    query: Tuple[str, ...], start: datetime, end: datetime, period: str, hidden: bool
):
    """Generate a graph displaying the number of created/open/closed tasks over time"""

    if start >= end:
        click.echo(error(f"Start date must be before end date {start} >= {end}"))
        return 1

    with orm.db_session:
        spinner = Halo(text="Constructing report")
        spinner.start()
        tasks = list(
            taskQuery(
                lambda t: (hidden or t.hidden == False)
                and (t.createdDate >= start or t.doneDate >= start)
                and (t.createdDate <= end or t.doneDate <= end),
                *query,
            )
        )

        keys: List[str] = []
        data: Dict[str, List[int]] = {}

        multi: Dict[str, bool] = {
            "year": start.year != end.year,
            "month": start.month != end.month,
        }
        if period == "year":
            diff = timedelta(days=365)
        elif period == "month":
            diff = timedelta(days=30)
        elif period == "week":
            diff = timedelta(days=7)
        else:
            diff = timedelta(days=1)

        if period == "year":
            fmt = "%Y"
        elif period == "month":
            fmt = "%b %Y" if multi["year"] else "%b"
        elif period == "week":
            fmt = "%Y W%W" if multi["year"] else "W%W"
        else:
            fmt = "%b %d %Y" if multi["year"] else ("%b %d" if multi["month"] else "%d")

        current = start
        while current <= end:
            key = current.strftime(fmt)
            data[key] = [0, 0, 0, 0]
            keys.append(key)
            current += diff
        key = end.strftime(fmt)
        if key not in data:
            data[key] = [0, 0, 0, 0]
            keys.append(key)

        for task in tasks:
            tstart = task.createdDate if task.createdDate >= start else start
            tend = task.doneDate if task.doneDate else end

            istart = keys.index(tstart.strftime(fmt))
            iend = keys.index(tend.strftime(fmt))

            data[keys[istart]][0] += 1
            if task.doneDate is not None:
                data[keys[iend]][3] += 1

            if istart != iend:
                for k in keys[istart + 1 : iend + 1]:
                    if any(
                        [
                            datetime.fromtimestamp(x[0]).strftime(fmt) == k
                            for x in task.events
                        ]
                    ):
                        data[k][2] += 1
                    else:
                        data[k][1] += 1

        maximum_sum = max([sum(x) for x in data.values()])

        def bar(new: int, open: int, active: int, closed: int) -> str:
            sum = new + open + closed + active
            if sum == 0:
                return ""
            size = int(__BAR_WIDTH__ * sum / maximum_sum)
            ret = ""
            for (val, fg) in [
                (new, "yellow"),
                (open, "bright_black"),
                (active, "blue"),
                (closed, "green"),
            ]:
                if sum == 0:
                    break
                comp = int(size * val / sum)
                ret += click.style(__BAR_CHAR__ * comp, fg=fg)
                size -= comp
                sum -= val
            return ret

        table = Table(align=[">", ">", ">", ">", ">", "<"])
        for key in keys:
            row = data[key]
            table.append(
                [
                    click.style(key, bold=True),
                    click.style(str(row[0]), fg="yellow", bold=True),
                    click.style(str(row[1]), fg="bright_black", bold=True),
                    click.style(str(row[2]), fg="blue", bold=True),
                    click.style(str(row[3]), fg="green", bold=True),
                    bar(*row),
                ]
            )

        spinner.stop()

        click.echo(table.format())



from typing import Dict
import uuid
from hashlib import md5
from datetime import datetime, timedelta
from pony.orm import *

db = Database()


class Tag(db.Entity):
    uuid = PrimaryKey(uuid.UUID, default=uuid.uuid4)
    label = Required(str)
    tasks = Set("Task")


class Task(db.Entity):
    uuid = PrimaryKey(uuid.UUID, default=uuid.uuid4)
    state = Required(str, default="OPEN")
    hidden = Required(bool, default=False)
    label = Required(str)
    priority = Required(int, default=5)
    tags = Set("Tag")
    parents = Set("Task", reverse="children")
    children = Set("Task", reverse="parents")
    dueDate = Optional(datetime)
    goalDuration = Optional(timedelta)
    goalPeriod = Optional(timedelta)
    createdDate = Required(datetime, default=datetime.now())
    updatedDate = Required(datetime, default=datetime.now())
    doneDate = Optional(datetime)
    events = Required(Json, default=[])

    @property
    def shortUuid(self) -> str:
        return str(self.uuid).split("-", 1)[0]

    @property
    def active(self) -> bool:
        return len(self.events) != 0 and len(self.events[-1]) == 1

    @property
    def urgency(self) -> float:
        urg = 5 - self.priority
        urg += 0.1 * (datetime.now() - self.createdDate).total_seconds() / 86400
        if self.active:
            urg += 2.0
        if self.dueDate:
            urg += 8.0 * (
                (datetime.now() - self.createdDate) / (self.dueDate - self.createdDate)
            )

        if self.goalDuration and self.goalPeriod:
            time = 0.0
            now = datetime.now().timestamp()
            P = self.goalPeriod.total_seconds()
            for event in [x for x in self.events if len(x) == 1 or x[1] > (now - P)]:
                a = max(event[0], now - P) - now
                b = (event[1] if len(event) == 2 else now) - now
                t = (b - a) * (2 * P + b + a) / (2 * P)
                time += t

            urg += 8.0 * max(1 - (time / self.goalDuration.total_seconds()), 0)
        return urg

    @property
    def priorityColor(self) -> Dict:
        pri = self.priority
        if pri <= 0:
            return {"fg": "red", "bold": True}
        elif pri <= 1:
            return {"fg": "red"}
        elif pri <= 2:
            return {"fg": "yellow"}
        elif pri <= 3:
            return {"fg": "green"}
        elif pri <= 4:
            return {"fg": "blue"}
        return {}

    @property
    def urgencyColor(self) -> Dict:
        urg = self.urgency
        if urg >= 10.0:
            return {"fg": "red", "underline": True, "bold": True}
        elif urg >= 9.0:
            return {"fg": "red", "bold": True}
        elif urg >= 8.0:
            return {"fg": "red"}
        elif urg >= 6.0:
            return {"fg": "yellow"}
        elif urg >= 4.0:
            return {"fg": "green"}
        elif urg >= 1.0:
            return {"fg": "blue"}
        return {}

    @property
    def color(self) -> str:
        return f"#{int(md5((self.label + self.createdDate.isoformat()).encode('utf-8')).hexdigest(), 16) % 0xFFFFFF:06X}"
